#define LUMINANCE_VECTOR float3(0.2125, 0.7154, 0.0721)
#define BUCKET_SIZE 64
#define minLum2 12
#define maxLum2Inv (1.0f/14)

groupshared uint sharedGroup[BUCKET_SIZE];

Texture2D<float4> backbuffer : register(t0);

// size [1][1]
// GFXFormatR16F
RWTexture2D<float4> lumOut : register(u0);

// size [64][1]
// GFXFormatR32U
// Histogram
RWTexture1D<uint> bucketTexture : register(u1);

RWTexture2D<float4> histogramTexture : register(u2);

cbuffer RuntimeSettingsCBuffer : register(b0)
{
    float deltaRealTime;
    uint2 destinationDimensions;
};

float hdrLuminance(float3 sample) {
    return dot(sample, LUMINANCE_VECTOR);
}

uint getHistogramBin(float lum) {
    return saturate((log2(lum + 1e-10) + minLum2) * maxLum2Inv) * (BUCKET_SIZE -1);
}

float getLuminance(uint bin) {
    return exp2(((bin / float(BUCKET_SIZE -1)) / maxLum2Inv) - minLum2);
}

float getWeightFactor(float2 uv) {
    // vec2 ndc = uv * float2(2.0f, 2.0f) - float2(1.0f, 1.0f);
    float2 ndc = uv * 2.0f - 1.0f;
    float sqrDist = dot(ndc, ndc);
    float factor = saturate(1.0f - sqrDist);
    return factor * factor;
}

float getWeightFactor_old(float2 uv) {
    float dotNormalized = (dot(uv - 0.5f, uv - 0.5f) * 2);
    return saturate(1.0 - pow(dotNormalized, 2) * 10);
    // return saturate(1.0 - pow(dotNormalized, 4) * 25);
}

// Define the number of threads to be executed in a single thread group when we dispatch the compute shader.
[numthreads(8, 8, 1)]
void MAIN0(uint3 dispatchThreadID : SV_DispatchThreadID, uint3 groupThreadId : SV_GroupThreadID)
{
    uint sharedGroupPos = groupThreadId.y * 8 + groupThreadId.x;
    sharedGroup[sharedGroupPos] = 0;

    GroupMemoryBarrierWithGroupSync();

    const uint2 storePos = uint2(dispatchThreadID.xy);
    const float4 sample = backbuffer[storePos];

    const float lum = hdrLuminance(sample.xyz);
    const uint bin = getHistogramBin(lum);

    const float2 uv = (dispatchThreadID.xy + 0.5) * (1.0 / destinationDimensions.xy);
    const uint weight = (uint)(64.0 * getWeightFactor(uv));

    InterlockedAdd(sharedGroup[bin], weight);

    GroupMemoryBarrierWithGroupSync();

    InterlockedAdd(bucketTexture[sharedGroupPos], sharedGroup[sharedGroupPos]);
}

[numthreads(1, 1, 1)]
void MAIN1(uint3 dispatchThreadID : SV_DispatchThreadID)
{
    float totalLum = 0.0f;
    float totalWeight = 0.0f;
    if (dispatchThreadID.x == 0 && dispatchThreadID.y == 0) {
        float prevLuma = lumOut[uint2(0,0)].x;
        // if (prevLuma == 0 || isnan(prevLuma)) {
        if (isnan(prevLuma)) {
            prevLuma = 1;
        }

        for (uint i = 0; i < BUCKET_SIZE; i++) {
            float binWeight = bucketTexture[i];
            totalLum += getLuminance(i) * binWeight;

            totalWeight += binWeight;
            bucketTexture[i] = 0;
        }
        float targetLuma = (totalLum / totalWeight);

        float diff = targetLuma - prevLuma;
        // TODO: This isn't particulary good since the fewer fps the user has the longer the interpolation will be.
        float newAdaptedLuma = prevLuma + ( diff * ( 1.0 - exp( -deltaRealTime * 1.1f ) ) );

        lumOut[uint2(0, 0)] = float4(newAdaptedLuma, 0, 0, 1);
    }
}

[numthreads(1, 1, 1)]
void MAIN2(uint3 dispatchThreadID : SV_DispatchThreadID)
{
    float maxValue = 0.0f;
    if (dispatchThreadID.x == 0 && dispatchThreadID.y == 0) {
        for (uint i = 0; i < BUCKET_SIZE; i++) {
            maxValue = max(bucketTexture[i], maxValue);
        }

        for (uint x = 0; x < BUCKET_SIZE; x++) {
            float val = bucketTexture[x] / maxValue * BUCKET_SIZE;
            for (uint y = 0; y < BUCKET_SIZE; y++) {
                if (val > y) {
                    histogramTexture[uint2(x, (BUCKET_SIZE - 1) - y)] = float4(1, 1, 1, 1);
                } else {
                    histogramTexture[uint2(x, (BUCKET_SIZE - 1) - y)] = float4(0, 0, 0, 1);
                }
            }
            bucketTexture[x] = 0;
        }
    }
}